#ifndef MSGNO_H
#define MSGNO_H

/* msgno.h
 * Managing Error Codes and Associated Messages Across Separate C Libraries
 * Copyright (c) 2001 Michael B. Allen <mballen@erols.com>
 *
 * The MIT License
 * 
 * Permission is hereby granted, free of charge, to any person obtaining a
 * copy of this software and associated documentation files (the "Software"),
 * to deal in the Software without restriction, including without limitation
 * the rights to use, copy, modify, merge, publish, distribute, sublicense,
 * and/or sell copies of the Software, and to permit persons to whom the
 * Software is furnished to do so, subject to the following conditions:
 * 
 * The above copyright notice and this permission notice shall be included
 * in all copies or substantial portions of the Software.
 * 
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL
 * THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR
 * OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE,
 * ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR
 * OTHER DEALINGS IN THE SOFTWARE.
 */ 

#ifndef MSGNO

#define MSG(fmt, args...)
#define MNO(msgno)
#define MNF(msgno, fmt, args...)
#define PMSG(fmt, args...)
#define PMNO(msgno)
#define PMNF(msgno, fmt, args...)
#define AMSG(fmt, args...)
#define AMNO(msgno)
#define AMNF(msgno, fmt, args...)

#else

#if __GNUC__ <= 2 && __GNUC_MINOR__ < 96

#define MSG(fmt, args...) _msgno_printf("%s:%u: " fmt "\n", \
					__FILE__, __LINE__ , ## args)
#define MNO(msgno) _msgno_printf("%s:%u:%s: %s\n", \
					__FILE__, __LINE__, msgno_msg(msgno))
#define MNF(msgno, fmt, args...) _msgno_printf("%s:%u: %s" fmt "\n", \
					__FILE__, __LINE__, msgno_msg(msgno) , ## args)

#define PMSG(fmt, args...) (_msgno_buf_idx = sprintf(_msgno_buf, \
					"%s:%u: " fmt "\n", __FILE__, __LINE__ , ## args))
#define PMNO(msgno) (_msgno_buf_idx = sprintf(_msgno_buf, \
					"%s:%u: %s\n", __FILE__, __LINE__, msgno_msg(msgno)))
#define PMNF(msgno, fmt, args...) (_msgno_buf_idx = sprintf(_msgno_buf, \
					"%s:%u: %s" fmt "\n", __FILE__, __LINE__, msgno_msg(msgno) , ## args))

#define AMSG(fmt, args...) (_msgno_buf_idx += sprintf(_msgno_buf + _msgno_buf_idx, \
					"  %s:%u: "fmt"\n", __FILE__, __LINE__ , ## args))
#define AMNO(msgno) (_msgno_buf_idx += sprintf(_msgno_buf + _msgno_buf_idx, \
					"  %s:%u: %s\n", __FILE__, __LINE__, msgno_msg(msgno)))
#define AMNF(msgno, fmt, args...) (_msgno_buf_idx += sprintf(_msgno_buf + _msgno_buf_idx, \
					"  %s:%u: %s" fmt "\n", __FILE__, __LINE__, msgno_msg(msgno) , ## args))

#else

#define MSG(fmt, args...) _msgno_printf("%s:%u:%s: " fmt "\n", \
					__FILE__, __LINE__, __FUNCTION__, ## args)
#define MNO(msgno) _msgno_printf("%s:%u:%s: %s\n", \
					__FILE__, __LINE__, __FUNCTION__, msgno_msg(msgno))
#define MNF(msgno, fmt, args...) _msgno_printf("%s:%u:%s: %s" fmt "\n", \
					__FILE__, __LINE__, __FUNCTION__, msgno_msg(msgno), ## args)

#define PMSG(fmt, args...) (_msgno_buf_idx = sprintf(_msgno_buf, \
					"%s:%u:%s: " fmt "\n", \
					__FILE__, __LINE__, __FUNCTION__, ## args))
#define PMNO(msgno) (_msgno_buf_idx = sprintf(_msgno_buf, \
					"%s:%u:%s: %s\n", \
					__FILE__, __LINE__, __FUNCTION__, msgno_msg(msgno)))
#define PMNF(msgno, fmt, args...) (_msgno_buf_idx = sprintf(_msgno_buf, \
					"%s:%u:%s: %s" fmt "\n", \
					__FILE__, __LINE__, __FUNCTION__, msgno_msg(msgno), ## args))

#define AMSG(fmt, args...) (_msgno_buf_idx += sprintf(_msgno_buf + _msgno_buf_idx, \
					"  %s:%u:%s: " fmt "\n", \
					__FILE__, __LINE__, __FUNCTION__, ## args))
#define AMNO(msgno) (_msgno_buf_idx += sprintf(_msgno_buf + _msgno_buf_idx, \
					"  %s:%u:%s: %s\n", \
					__FILE__, __LINE__, __FUNCTION__, msgno_msg(msgno)))
#define AMNF(msgno, fmt, args...) (_msgno_buf_idx += sprintf(_msgno_buf + _msgno_buf_idx, \
					"  %s:%u:%s: %s" fmt "\n", \
					__FILE__, __LINE__, __FUNCTION__, msgno_msg(msgno), ## args))

#endif
#endif

#define NULL_POINTER_ERR _builtin_codes[0].msgno

struct msgno_entry {
	unsigned int msgno;
	const char *msg;
};

extern struct msgno_entry _builtin_codes[];

int msgno_add_codes(struct msgno_entry *list);
const char *msgno_msg(unsigned int errnum);

int msgno_hdlr_stderr(const char *fmt, ...);
extern int (*msgno_hdlr)(const char *fmt, ...);

extern char _msgno_buf[];
extern unsigned int _msgno_buf_idx;
void _msgno_printf(const char *fmt, ...);

#endif /* MSGNO_H */
